; ints module - "Nice" interrupt handlers
;
; 1. Exception handlers
; 2. Auto-generate default interrupt handlers. These tell you what
;    interrupt number it is, as a debugging aid.
;
; Notes:
;	Exception 8 (double fault) and 10 (invalid TSS) should have task
;	gates, so a known machine state is entered when these exceptions are
;	triggered.
;

	SECTION .text
init_ints:
;	call	generate	;Done in init.inc!
	callproc patch_int,	6, exc6_handler
	callproc patch_int,	12, exc12_handler
	callproc patch_int,	13, exc13_handler
	ret



%if 0
CODE_end	equ	0x60000

generate:
	pusha
	mov	edi, IDT
	mov	ebx, [CODE_end]
	mov	ecx, 48		; Generate handlers for first 48 interrupts
	xor	edx, edx	; interrupt #
.loop
	;; Generate IDT entry
	mov	eax, ebx
	mov	[edi], ax
	shr	eax, 16
	mov	[edi+2], word 8	; flat_code selector
	mov	[edi+4], word 0x8E00; flags: INT and PRESENT
	mov	[edi+6], ax
	add	edi, 8

	;; Generate machine code for interrupt handler
	mov	[ebx], byte 0x68; push (interrupt #)
	mov	[ebx+1], edx
	inc	edx
	mov	[ebx+5], byte 0xE9; jmp default_int_handler
	mov	eax, default_int_handler; compute offset
	sub	eax, ebx
	sub	eax, 10
	mov	[ebx+6], eax
	add	ebx, 10		; 10 bytes of machine code
	loop	.loop
	mov	[CODE_end], ebx
	popa
	ret
%endif


default_int_handler:
	sprint	10," * Interrupt "
;	sub	ebp, 4
;	mpush	[esp]
;	mpop	S(0)
;	call	print_hex
	cr
	add	esp, 4
	iretd



exc6_handler:
	sprint	10," * EXC6: Invalid opcode",10
	mov	eax, [esp]
	DPUSH	eax
	call	print_hex
	cr
	sti
	jmp short $

exc12_handler:
	sprint	10," * EXC12: Stack fault",10
	sti
	jmp short $

exc13_handler:
	sprint	10," * EXC13: GPF",10
	DPUSH	eax
	mpush	[esp]
	mpop	S(0)
	call	print_hex
	DPUSH	eax
	mpush	[esp+4]
	mpop	S(0)
	call	print_hex
	cr
	sti
	jmp short $
