; loader.inc - New RDF loader, which uses the ODB
;
; Copyright (C) 1999 Thomas Novelli
;	Retro is distributed under the terms of the GNU LGPL, version 2 or
;	greater. (For the text of the license, see the file COPYING.LIB)
;
; Bugs:
;	4/9/99	(Not really a bug): I had to put .bss in CODE space, because
;		the BSS_reserve doesn't come until the _end_ of the RDF
;		module, which was causing the object-tree nodes to go in
;		.bss space, where they promptly get wiped out when the
;		module executes. I mention this here because we can't do
;		this if we ever want to make code space read-only.
;
; To do:
;	Use the 'library' directive as a kludge to specify module names,
;	where to load/search in the hierarchy, etc.  By default, objects
;	go under 'kern', just like they used to.
;
;		Example:
;		library "=module_name"
;		library "^parent_name"
;
;	Align .text/.data/.bss segments on a 16-byte boundary
;
;	Save the RDF header records under a child object with a special
;	name. (for re-linking)  (should we bother now?)
;
; There will be a "full version" of the loader, so this one doesn't have to
; be complete. Besides, the full one will handle XCOM, Retro/Tunes native
; format, maybe ELF and/or A.OUT too...
;


load:
;
; Input (tentative):
%define object		ebp+4*4	;Address of object to load
%define parent		ebp+3*4	;OID of object's parent-to-be
%define flags		ebp+2*4	;flags to put in ODB entry
;
; Output (tentative):
%define r_next		ebp+4*4	;Address at end of module (next object)
%define code		ebp+3*4	;Address of .text segment
%define result		ebp+2*4	;Result code
;
; Local Variables (remember to allocate enough space! SUB ESP, n*4)
%define obj_len		ebp-1*4	;Length of module
%define bss_len		ebp-2*4	;Length of BSS segment
%define next		ebp-3*4	;Next object
%define hdr_end		ebp-4*4	;End of RDF header
;
;
%define seg_addr	0x70000	;.text/.data/.bss addresses, and an address
				;for every 'import' as well
;
;Questions:
; Should we run objects automatically, depending on flags?
; Doesn't look like that's possible w/ RDF. Just run first thing in .text.
;
	push	ebp
	mov	ebp, esp
	sub	esp, 4*4	;Allocate local variables
	pusha

	callproc print,	.msg

	mov dword [result], 1	;Set error flag

	;; Check "magic number" ("RDOFF2")
	mov	esi, [object]
	mov	edi, MAGIC
	mov	ecx, 6
	cld
	rep cmpsb
	jz	.ok
;	callproc print,	err_msg		;Error - not RDF 2.0
	mov dword [result], 2
	jmp	.end
.ok

;ESI = [object] + 6
	mov	eax, [esi]		;Length
	mov	[obj_len], eax
	lea	eax, [esi+4+eax]	;Next object
	mov	[next], eax
	mov	ebx, [esi+4]		;Header length
	add	esi, 8
	mov	[hdr_end], ebx		;code (.text segment) address
	add	[hdr_end], esi
;ESI = start of header

	;; Read segment headers (only 2) & copy each space to new location
	push	esi
	add	ebx, esi	;Skip past header
.seg_hdr
	mov	ax, [ebx]		;seg. type
	or	ax, ax			;null segment? done.
	jz	.seg_done
	movzx	edx, word [ebx+2]	;edx = seg. #
	mov	ecx, [ebx+6]		;seg. length
	add	ebx, byte 10		;skip seg. header
	;; Determine new location:
	;;  For .text segment (#0), move it to CODE_end
	;;  For .data segment (#1), move it to DATA_end
	or	edx, edx	;Segment #0 (.text)?
	jnz	.get_data
	mov	edi, [CODE_end]		;edi = new location
	jmp short .get_code
.get_data
	mov	edi, [DATA_end]
.get_code

	mov	esi, ebx		;esi = old location
	add	ebx, ecx	;ebx = next seg.

	mov	[seg_addr + edx*4], edi	;save address..
	rep	movsb			;Copy to new location

	;Update free space ptr (CODE/DATA_end)
	or	edx, edx	;Segment #0 (.text)?
	jnz	.set_data
	mov	[CODE_end], edi		;edi = new location
	jmp short .set_code
.set_data
	mov	[DATA_end], edi
.set_code
	jmp short .seg_hdr	;Next segment...
.seg_done
	mov	ebx, [CODE_end]		;Get an address for .BSS
	mov	[seg_addr + 8], ebx
	pop	esi

	;Read & process the RDF header
.header
	movzx	edx, byte [esi]		;record type
	cmp	dl, 5			;Range checking.. avoid bad CALLs.
	ja	.hdr_skip
	shl	edx, 2
	xor	edi, edi	;Clear error flag
	push	esi
	call	[rec_handlers + edx]	;Call via address table
	pop	esi
	or	edi, edi	;Error?
	jnz	.hdr_error
.hdr_skip
	movzx	eax, byte [esi+1]
	add	esi, eax		;Skip to next record
	add	esi, 2
	cmp	esi, [hdr_end]		;Done yet?
	jb	.header

	;Successful completion
	mov dword [result], 0	;Reset error flag
	mov	eax, [seg_addr+0]	;Address of .text
	mov	[code], eax
	mov	eax, [next]		;Address of next object
	mov	[r_next], eax
.end
	popa
	leave
	ret
.hdr_error
	callproc print,	.errmsg
	mov	[result], edi
	jmp short .end

.msg		db "Loading... ",0
.errmsg		db "load: header error",10,0



rec_handlers:
;
; Table of record handler routines.
;
; Input:
;	ESI = beginning of record
;	EDI = 0 (So you don't need to change it unless there's an error)
; Output:
;	EDI = error # (0 if successful)
;
; Feel free to change EAX, EBX, ECX, EDX, ESI, EDI
;
	dd	rec_null
	dd	rec_reloc
	dd	rec_import
	dd	rec_export
	dd	rec_library
	dd	rec_bss



rec_null:
	callproc print_hex,	esi
	callproc print,		.start_msg
	mov	edi, 1	; set error flag
	ret
.start_msg	db	"_load: RDF type 0 record found (shouldn't happen!).",10,0



rec_reloc:
;	callproc print, .start_msg

	movzx	eax, word [esi+8]	;get segment # to link to
	shl	eax, 2
	mov	eax, [seg_addr + eax]	;EAX = base address of segment

	movzx	ebx, byte [esi+2]		;get segment #
	and	bl, 0xBF		;mask off bit 6
	shl	ebx, 2
	mov	ebx, [seg_addr + ebx]
	mov	edx, ebx		;EBX = Address of this segment
	add	ebx, [esi+3]		;EBX = Address requiring reloc

	test	byte [esi+2], 0x40	;Relative relocation?
	jz	.normal
	sub	eax, edx		;Just add the difference between
					;this segment and the destination
.normal
	mov	cl, [esi+7]		;CL = Length of item (1/2/4 bytes)
	cmp	cl, 1
	je	.byte
	cmp	cl, 2
	je	.word
.dword
	add	[ebx], eax
	ret
.word
	add	[ebx], ax
	ret
.byte
	add	[ebx], al
	ret
;.start_msg	db	"Reloc ",0



rec_import:
;	callproc print, .start_msg
	add	esi, byte 4		;esi = start of name string
	;;	name, length=unknown, top, ifs=0
	mpush	esi, -1, [parent], 0
	call	search
	pop	edx		;address
	add	esp, 3*4

	or	edx, edx
	jz	.error

	;Found it
	movzx	eax, word [esi-2]	;Get seg. #
	mov	[seg_addr + eax*4], edx	;Store addr in seg. address table
	xor	edi, edi	;Clear error (since we changed EDI)
	ret
.error
	mov	edi, 2
	callproc print,	.error_msg
	callproc print,	esi
	cr
	ret
;.start_msg	db	"Import ",0
.error_msg	db	"RDF Import: can't find ",0



rec_export:
;	callproc print, .start_msg

	;; Get address of object
	movzx	edx, byte [esi+2]	;segment #
	mov	eax, [seg_addr + edx*4]	;get seg address
	add	eax, [esi+3]		;add offset
	;; eax = address

	;; Get type (code or data)
	cmp	dl, 2	;.bss is data
	je	.bss
	inc	edx			;Type 1 (T_FN) if in .text
.bss					;Type 2 (T_BIN) if in .data/.bss
	;; edx = flags

	mov	edi, [DATA_end]
	push	edi
	;; edi = name ptr.

	;; Copy name string to DATA area
	movzx	ecx, byte [esi+1]
	sub	cl, 5
	add	[DATA_end], ecx		;allocate space
	add	esi, 7
	rep	movsb			;copy it

	pop	edi
	callproc insert, edi, [parent], eax, 0, edx

	xor	edi, edi	;Clear error flag
	ret
;.start_msg	db	"Export ",0



rec_library:
;;
;;
;;
	callproc print, .start_msg
	callproc print, [esi+2]
	cr
	ret
.start_msg	db	"Library: ",0



rec_bss:
;	callproc print, .start_msg
	mov	eax, [esi+2]		;EAX = amount to reserve
	add	[CODE_end], eax
	ret
;.start_msg	db	"BSS ",0



; Constants
;err_msg:	db	"load: Not a RDOFF 2.0 object!",10,0
MAGIC:		db	"RDOFF2"
	db 0
