; 5/17/99 version

extern readkey,getc,putc,read_com_byte,write_com_byte
extern send_break,set_com_speed,set_rx_handler,set_error_handler
extern set_com_bits,get_RTS_DTR,set_RTS,set_DTR,reset_RTS,reset_DTR
extern print,print_hex,print_num,debug

%include "boot.mac"
%include "keydefs.mac"

init:
  mov al,0
  mov [Show_Stuff],al
  mov al,0
  mov ebx,So_we_received_something
  call set_rx_handler
  mov al,1
  mov ebx,So_we_received_something
  call set_rx_handler
  mov al,1
  mov ebx,Error_Handler
  call set_error_handler
  mov al,0
  mov ebx,Error_Handler
  call set_error_handler
  call Reset_nbr_errors
  ret

global terminal
terminal:
    pushad ; It seems forth does not like I change ESI
    mov byte[CommPort],0 ; Init comport
    mov byte[Show_Stuff],1
    mov dword[speed],speeds
    mov dword[current_nbr_bits],nbr_bits
    callproc print,init_msg
  .DoReadkey:
    call readkey
    ; now ah=scan-code al=ascii, ax=0 if no key
    or ax,ax
    jz .Received?
  .there_is_something:
    cmp ah,KEY_F1
    jb .return_key?
    cmp ah,KEY_F10 ; ignore F11 and F12 because F11 scancode not F10+1
    ja .return_key?
    ;callproc putc,'&'
    ; So it is a F1-F10 key
    sub ah,59 ; 59=F1
    shr eax,6 ; al should now be ah*4, once we'll mask 'bad' bits
    and eax,00000000000000000000001111111100b
    call [eax+fctns_keys]
    jmp .Received?
  .return_key?:
    and ax,0ffh
    cmp al,0ah
    jnz .Escape?
    mov al,0dh
    jmp .SendIt
  .Escape?:
    cmp al,27
    jnz .PlainASCII?
    jmp .quit
  .PlainASCII?: ; I don't want to send Shift, Control, ...
    cmp al,' '
    jb .Received?
    cmp al,127
    ja .Received?
  .SendIt:
    xchg ah,al
    mov al,byte[CommPort] ; com
    call write_com_byte
  .Received?:
    test byte[received_car],1
    jz .Redo_it
    call Receive_car
  .Redo_it:
    jmp .DoReadkey
  .quit:
    mov byte[Show_Stuff],0
    cr
    popad
    ret

Receive_car:
    mov byte[received_car],0
  .again:
    cmp byte[Show_Stuff],1
    jnz .nothing
    mov al,byte[CommPort]
    call read_com_byte
    jc .nothing
    xchg ah,al
    cmp al,0dh
    je .jmp_again
    push eax
    call putc
    pop eax
  .jmp_again:
    jmp .again
  .nothing:
    ret

Help:
    callproc print,help_msg
    ret

SetPort:
    ; On F2, change communication port (just 0 and 1 for now)
    inc byte[CommPort]
    mov al,byte[CommPort]
    cmp al,2
    jb .valid_comport
    mov byte[CommPort],0
  .valid_comport:
    callproc print,port_msg
    mov al,[CommPort]
    add al,'0'
    and al,0ffh
    callproc putc,eax
    cr
    ret

SetSpeed:
  .increment_speed:
    inc dword [speed]
    inc dword [speed] ; each speed is a word (2 bytes)
    mov eax,[speed]
    mov ax,[eax]
    and eax,0ffffh
    cmp ax,0
    jnz .valid_speed
    mov dword[speed],speeds ; let's return to smallest speed
    mov eax,[speed]
    mov ax,[eax]
    and eax,0ffffh
  .valid_speed:
    mov ebx,eax
    push ebx
    mov al,[CommPort]
    call set_com_speed
    callproc print, speed_msg
    pop ebx
    callproc print_num,ebx
    ret

SetBits:
  .increment_nbr_bits:
    inc dword [current_nbr_bits]
    mov eax,[current_nbr_bits]
    mov al,[eax]
    and eax,0ffh
    cmp ax,0
    jnz .request_this_nbr_bits
    mov dword[current_nbr_bits],nbr_bits ; let's return to smallest speed
    mov eax,[current_nbr_bits]
    mov al,[eax]
    and eax,0ffh
  .request_this_nbr_bits:
    mov ebx,eax
    mov ah,al
    push ebx
    mov al,[CommPort]
    call set_com_bits
    callproc print, bits_msg
    pop ebx
    callproc print_num,ebx
    ret

Toggle_RTS:
    mov al,[CommPort]
    call get_RTS_DTR
    test bl,2
    jz .Do_Set_RTS
    call reset_RTS
    jmp .end
  .Do_Set_RTS:
    call set_RTS
  .end:
    ret

Toggle_DTR:
    mov al,[CommPort]
    call get_RTS_DTR
    test bl,1
    jz .Do_Set_DTR
    call reset_DTR
    jmp .end
  .Do_Set_DTR:
    call set_DTR
  .end:
    ret

SendBreak:
    mov al,[CommPort]
    call send_break
    ret

ShowErrorReport:
    callproc print,overrun_err_msg
    xor eax,eax
    mov ax,[nbr_overrun_errors]
    callproc print_num,eax
    callproc print,parity_err_msg
    xor eax,eax
    mov ax,[nbr_parity_errors]
    callproc print_num,eax
    callproc print,framing_err_msg
    xor eax,eax
    mov ax,[nbr_framing_errors]
    callproc print_num,eax
    callproc print,nbr_errors_reset_msg
    cr
    call Reset_nbr_errors
    ret

Undefined:
    ret

; Called by the serial handler when rx_triger_level reached
So_we_received_something:
    mov byte[received_car],1
    ret

Reset_nbr_errors:
    xor eax,eax
    mov [nbr_overrun_errors],ax
    mov [nbr_parity_errors],ax
    mov [nbr_framing_errors],ax
    ret

Error_Handler:
  .Overrun?:
    test al,2
    jnz .Parity?
    inc word [nbr_overrun_errors]
  .Parity?:
    test al,4
    jnz .Framing?
    inc word [nbr_parity_errors]
  .Framing?:
    test al,8
    jnz .No_Error
    inc word [nbr_framing_errors]
  .No_Error:
    ret

SECTION .data

fctns_keys dd Help ; F1
           dd SetPort ; F2
           dd SetSpeed; F3
           dd SetBits; F4
           dd Undefined; F5
           dd Undefined; F6
           dd Toggle_RTS; F7
           dd Toggle_DTR; F8
           dd SendBreak; F9
           dd ShowErrorReport; F10

init_msg db 10,"Welcome to the serial port terminal, press F1 for help.",10,0
help_msg db 10,"F1: Help"
         db 10,"F2: Select comm. port"
         db 10,"F3: Select comm. speed"
         db 10,"F4: Select number of bits"
         db 10,"F7: Toggle RTS"
         db 10,"F8: Toggle DTR"
         db 10,"F9: Send a break signal"
         db 10,"F10: Show a report on errors since last time you check"
         db 10,"ESC return to DDR"
         db 10,0
port_msg: db 10,"current port is com",0
speed_msg: db 10,"new speed set to ",0
bits_msg: db 10,"number of bits set to ",0
speeds dw 1200,2400,9600,19200,38400,57600,0 ; a bug with 115200!
nbr_bits db 5,6,7,8,0
overrun_err_msg db      10,"overrun errors: ",0
parity_err_msg db       10,"parity errors:  ",0
framing_err_msg db      10,"framing errors: ",0
nbr_errors_reset_msg db 10," resetting error counts",10,0

SECTION .bss
Show_Stuff resb 1
CommPort resb 1
speed resd 1 ; pointer amongs speeds
current_nbr_bits resd 1 ; pointer amongs nbr_bits
received_car resb 1
nbr_overrun_errors resw 1
nbr_parity_errors resw 1
nbr_framing_errors resw 1
