LOAD    ;( i*x u -- j*x , Interpret from disk block u )
        push    ecx
        push    esi
        push    cell [_BLK]
        push    cell [_SOURCE_ID]
        push    cell [_INBUF]
        push    cell [_INBUF_LEN]
        push    cell [_TO_IN]

        tos     ecx
        mov     [_BLK], ecx
        mov     [_SOURCE_ID], cell 0
        call    BLOCK
        upop    esi

        mov     ecx, 16
.loop
        mov     [_INBUF], esi
        mov     [_INBUF_LEN], cell 64
        mov     [_TO_IN], cell 0
        call    PARSE_INBUF
        add     esi, 64
        loop    .loop

        pop     cell [_TO_IN]
        pop     cell [_INBUF_LEN]
        pop     cell [_INBUF]
        pop     cell [_SOURCE_ID]
        pop     cell [_BLK]
        pop     esi
        pop     ecx
        ret
end LOAD



BLOCK   ;( u -- a-addr , Load disk block u; a-addr is the buffer )
        push    eax
        push    ebx
        push    ecx
        push    edx

        call    SEARCH_TABLE
        cmp     tosreg, cell 0
        jne     .end

        call    DROP
        tos     eax             ;AX = block #
        call    ALLOCATE_BUFFER

        mov     ecx, 5          ;# of retries (if error)
.read
        push    eax
        push    ecx
        call    CALC_SECTOR

        mov     ax, ds          ;ES=DS
        mov     es, ax
        tos     ebx             ;Address of buffer
        mov     ax, 0x0202      ;Service 2, read 2 sectors
        int16   0x13

        jnc     .ok             ;Check for errors

        mov     ah, 0           ;Reset disk
        int16   0x13
        pop     ecx
        pop     eax
        loop    .read

.ok
        pop     ecx
        pop     eax
.end
        pop     edx
        pop     ecx
        pop     ebx
        pop     eax
        ret

;
; Block-subsystem variables
;
disk
.sectors        dc      18
.tracks         dc      80
.heads          dc      2

buftbl  ;Buffer table:          Block#  Updated?        Disk#/File#  Priority
TIMES   NUM_BUFFERS     dc      0,      0
; Note: When adding more fields to the table, need to modify:
;       - Decrease the shift value for SHL ("Get buffer address")
;       - Increase the multiplier (currently 2)
;       (These are all in routines that deal with the table, namely
;       ALLOCATE_BUFFER and SEARCH_TABLE, also SAVE_BUFFER)

_buffers        dc      0x10000-(1024 * NUM_BUFFERS)    ;Address of buffers

_cur_buf        dc      0       ;Current buffer
_SCR            dc      0       ;Last block displayed on screen

end BLOCK



BUFFER  ;( u -- a-addr , a-addr is the buffer for disk block u )

        call    SEARCH_TABLE
        cmp     tosreg, cell 0
        jne     .end

        call    DROP
        call    ALLOCATE_BUFFER
.end
        ret
end BUFFER



ALLOCATE_BUFFER
;( u -- a-addr , Allocates a free buffer for block u )
;
; To-do: Priorities (use "oldest" buffer if one isn't free)
;
        push    eax
        push    ebx
        push    ecx

        mov     ebx, buftbl
        mov     ecx, NUM_BUFFERS
        xor     eax, eax                ;AX = unmodified buffer
.loop
        cmp     [ebx], cell 0
        je      .alloc
        cmp     [ebx+CELLSIZE], cell 0  ;Also look for unmodified buffers
        jne     .skip
        mov     eax, ebx
.skip
        add     ebx, 2*CELLSIZE
        loop    .loop

        ;If there's an unmodified buffer available, steal it
        cmp     eax, 0
        jne     .steal

        ;Otherwise, save a buffer, then steal it
        mov     ebx, buftbl             ;For now, just grab the first one
jmp hang
;Problem: This isn't working right. It's bungling the data, misplacing blocks..
        upsh    ebx
        call    SAVE_BUFFER
        jmp     .alloc

.steal
        mov     ebx, eax
.alloc
        upop    ecx                     ;Assign buffer to block u
        mov     [ebx], ecx
        mov     [_cur_buf], ebx         ;Set as current buffer
        sub     ebx, buftbl             ;Get buffer address
        shl     ebx, 8
        add     ebx, [_buffers]
        upsh    ebx

        pop     ecx
        pop     ebx
        pop     eax
        ret
end ALLOCATE_BUFFER



SAVE_BUFFER
;( a-addr -- , Save buffer whose table entry is at a-addr )
        push    eax
        push    ebx
        push    ecx
        push    edx

        upop    ebx
        mov     [ebx+CELLSIZE], cell 0  ;Mark as unmodified
        mov     eax, [ebx]
        upsh    eax
        call    CALC_SECTOR
        sub     ebx, buftbl             ;Get buffer address
        shl     ebx, 8
        add     ebx, [_buffers]

        mov     ax, ds          ;ES=DS
        mov     es, ax
        mov     ax, 0x0302      ;Service 3, write 2 sectors
        int16   0x13

        pop     edx
        pop     ecx
        pop     ebx
        pop     eax
        ret
end SAVE_BUFFER



SAVE_BUFFERS    ;( -- , Save all UPDATEd buffers )
        push    ebx
        push    ecx

        mov     ebx, buftbl
        mov     ecx, NUM_BUFFERS
.loop
        cmp     [ebx+CELLSIZE], cell 0
        je      .skip
        upsh    ebx
        call    SAVE_BUFFER
.skip
        add     ebx, 2*CELLSIZE
        loop    .loop

        pop     ecx
        pop     ebx
        ret
end SAVE_BUFFERS



SEARCH_TABLE
;
;( u -- u 0 | a-addr )
;
;If block u is in memory, return the buffer address. Otherwise, return 0.
;
        push    eax
        push    ebx
        push    ecx

        upop    eax
        mov     ebx, buftbl
        mov     ecx, NUM_BUFFERS
.loop
        cmp     [ebx], eax
        je      .found
        add     ebx, 2*CELLSIZE
        loop    .loop

        upsh    eax
        upsh    cell 0
        jmp     .end
.found
        mov     [_cur_buf], ebx         ;Set as current buffer
        sub     ebx, buftbl             ;Get buffer address
        shl     ebx, 8
        add     ebx, [_buffers]
        upsh    ebx
.end
        pop     ecx
        pop     ebx
        pop     eax
        ret
end SEARCH_TABLE



CALC_SECTOR
;
; Input: AX = Block number
; Output: in CX, DX
;
        dec     eax
        shl     eax, 1          ; Convert from blocks->sectors
        inc     eax

        xor     edx, edx
        div     cell [disk.sectors]
        mov     cl, dl

        xor     edx, edx
        div     cell [disk.heads]
        mov     dh, dl

        xor     edx, edx
        div     cell [disk.tracks]
        mov     ch, dl

        mov     dl, 0
        ret
end CALC_SECTOR



EMPTY_BUFFERS
;Simply zero-fill the entire buffer table
        push    ebx
        push    ecx

        mov     ebx, buftbl
        mov     ecx, NUM_BUFFERS * 2    ;(2 = # of cells per entry)
.loop
        mov     [ebx], cell 0
        add     ebx, CELLSIZE
        loop    .loop

        pop     ecx
        pop     ebx
        ret
end EMPTY_BUFFERS



FLUSH   ;( -- , Save & unassign all buffers )
        call    SAVE_BUFFERS
        call    EMPTY_BUFFERS
        ret
end FLUSH



UPDATE  ;( -- , Mark the current block as modified )
        push    ebx
        mov     ebx, [_cur_buf]
        mov     [ebx+CELLSIZE], cell -1
        pop     ebx
        ret
end UPDATE



BLK
        ret
end BLK



LIST    ;( u -- , Displays block u and sets SCR to u )
;
; List in 64x16 format, showing a blank for any unprintable characters
;
        push    eax
        push    ebx
        push    ecx

        call    CR

        tos     eax
        mov     [_SCR], eax

        call    BLOCK
        upop    ebx

        mov     ecx, 16
        xor     eax, eax
.l1
        push    ecx
        mov     ecx, 64
.l2
        mov     al, [ebx]
        cmp     al, 32
        jb      .nonprint
        cmp     al, 127
        ja      .nonprint
        jmp     .emit
.nonprint
        mov     al, 32
.emit
        upsh    eax
        call    EMIT
        inc     ebx
        loop    .l2

        call    CR
        pop     ecx
        loop    .l1

        pop     ecx
        pop     ebx
        pop     eax
        ret
end LIST


